import json
import logging
import ssl

import aiohttp

from aiohttp import web

from pypeman import endpoints, channels, nodes, message
from pypeman.errors import PypemanParamError


logger = logging.getLogger(__name__)


class HTTPEndpoint(endpoints.SocketEndpoint):
    """
    Endpoint to receive HTTP connection from outside.

    """
    def __init__(
            self,
            adress=None, address=None, port=None,  # obsolete params
            loop=None,
            http_args=None,
            host=None,
            sock=None,  # TODO: Why have sock and port if it's the same ?
            reuse_port=None,
            ):
        """
            :param http_args: dict to pass as **kwargs** to aiohttp.Application for example for
                `client_max_size`
            :param reuse_port: bool if true then the listening port specified in the url parameter)
                will be shared with other processes on same port
            :param host: string 'host:port' or ':port' or 'host'
            :param sock: host-string (same as host parameter)
                or socket-string ("unix:/sojet/file/path")
                or bound socket object
        """

        self.http_args = http_args or {}
        self.ssl_context = self.http_args.pop('ssl_context', None)
        self._app = None

        address = address or adress
        if address or port:
            raise PypemanParamError(
                "HTTPEndpoint Obsolete params ('address', 'port') ")

        if host and sock:
            raise PypemanParamError("There can only be one (parameter host or sock)")
        sock = sock or host or ''

        super().__init__(loop=loop, sock=sock, reuse_port=reuse_port)

    def add_route(self, *args, **kwargs):
        if self._app is None:
            self._app = web.Application(**self.http_args)

        self._app.router.add_route(*args, **kwargs)

    async def start(self):
        self.make_socket()
        if self._app is not None:
            srv = await self.loop.create_server(
                protocol_factory=self._app.make_handler(),
                sock=self.sock_obj,
                ssl=self.ssl_context,
            )
            message = "Server started at %r" % self.sock
            print(message)
            logger.info(message)
            return srv
        else:
            print("No HTTP route.")
            logger.warning("No HTTP route.")


class HttpChannel(channels.BaseChannel):
    """
    Channel that handles Http connection. The Http message is the message payload and some headers
    become metadata of message. Needs ``aiohttp`` python dependency to work.

    :param endpoint: HTTP endpoint used to get connections.
    :param method: Method filter.
    :param url: Only matching urls messages will be sent to this channel.
    :param encoding: Encoding of message. Default to 'utf-8'.

    """
    app = None

    def __init__(self, *args, endpoint=None, method='*', url='/', encoding=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.method = method
        self.url = url
        self.encoding = encoding
        if endpoint is None:
            raise TypeError('Missing "endpoint" argument')
        self.http_endpoint = endpoint

    async def start(self):
        first_start = self._first_start
        await super().start()
        if first_start:
            self.http_endpoint.add_route(self.method, self.url, self.handle_request)

    async def handle_request(self, request):
        content = await request.text()

        # extract match info from iohttp request
        meta = dict(request.match_info)

        # match infos will be overwritten if coinciding with known keywords like
        # 'url', 'method', . . .
        meta.update({
            'method': request.method,
            'url': str(request.url),
            'get_params': dict(request.query),
            })

        msg = message.Message(content_type='http_request', payload=content, meta=meta)
        try:
            result = await self.handle(msg)
            encoding = self.encoding or 'utf-8'
            return web.Response(body=result.payload.encode(encoding), status=result.meta.get('status', 200))

        except channels.Dropped:
            return web.Response(body="Dropped".encode('utf-8'), status=200)
        except Exception as e:
            return web.Response(body=str(e).encode('utf-8'), status=503)


class HttpRequest(nodes.BaseNode):
    """
        Http request node
        :param url: url to send.
        :param method: 'get', 'put' or 'post', use meta['method'] if None, Default to 'get'.
        :param headers: headers for request, use meta.get('headers') if None.
        :param cookies: cookies for request, use meta.get('cookies') if None.
        :param auth: tuple or aiohttp.BasicAuth object.
        :param verify: verify ssl. Default True.
        :param params: get params in dict. List for multiple elements, ex :
                       {'param1': 'omega', param2: ['alpha', 'beta']}
        :param client_cert: tuple with .crt and .key path
        :param binary: bool, Get response content as bytes
    """

    def __init__(self, url, *args, method=None, headers=None, auth=None,
                 verify=True, params=None, client_cert=None, cookies=None,
                 binary=False, json=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = url
        self.method = method
        self.headers = headers
        self.cookies = cookies
        self.auth = auth
        self.verify = verify
        self.params = params
        self.client_cert = client_cert
        self.url = self.url.replace('%(meta.', '%(')
        self.payload_in_url_dict = 'payload.' in self.url
        self.binary = binary
        self.json = json
        # TODO: create used payload keys for better perf of generate_request_url()

    def generate_request_url(self, msg):
        url_dict = msg.meta
        if self.payload_in_url_dict:
            url_dict = dict(url_dict)
            try:
                for key, val in msg.payload.items():
                    url_dict['payload.' + key] = val
            except AttributeError:
                self.channel.logger.exception(
                    "Payload must be a python dict if used to generate url. "
                    "This can be fixed using JsonToPython node before your "
                    "RequestNode")
                raise
        return self.url % url_dict

    async def handle_request(self, msg):
        """ generate url and handle request """
        url = self.generate_request_url(msg)
        loop = self.channel.loop
        conn = None
        ssl_context = None
        if self.client_cert:
            if self.verify:
                ssl_context = ssl.create_default_context()
            else:
                ssl_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
            try:
                ssl_context.load_cert_chain(self.client_cert[0], self.client_cert[1])
            except FileNotFoundError:
                logger.error("loading certs %s failed", self.client_cert)
                raise
            conn = aiohttp.TCPConnector(ssl=ssl_context, loop=loop)
        else:
            conn = aiohttp.TCPConnector(ssl=self.verify, loop=loop)

        headers = nodes.choose_first_not_none(self.headers, msg.meta.get('headers'))
        cookies = nodes.choose_first_not_none(self.cookies, msg.meta.get('cookies'))
        method = nodes.choose_first_not_none(self.method, msg.meta.get('method'), 'get')
        params = nodes.choose_first_not_none(self.params, msg.meta.get('params'))

        get_params = None
        if params:
            get_params = []
            for key, param in params.items():
                if isinstance(param, list):
                    for value in param:
                        get_params.append((key, value))
                else:
                    get_params.append((key, param))

        if isinstance(self.auth, tuple):
            basic_auth = aiohttp.BasicAuth(self.auth[0], self.auth[1])
        else:
            basic_auth = self.auth

        data = None
        if method.lower() in ['put', 'post']:
            data = msg.payload
        async with aiohttp.ClientSession(connector=conn, cookies=cookies) as session:
            resp = await session.request(
                    method=method,
                    url=url,
                    auth=basic_auth,
                    headers=headers,
                    params=get_params,
                    data=data
                    )
            if self.binary:
                resp_content = await resp.read()
            else:
                resp_content = str(await resp.text())
            if self.json:
                try:
                    resp_content = json.loads(resp_content)
                except Exception as exc:
                    logger.error(
                        "cannot json parse response from url %s (response=%r)",
                        url, resp_content)
                    raise exc
            return resp_content

    async def process(self, msg):
        """ handles request """
        msg.payload = await self.handle_request(msg)
        return msg



"""
This module contains all persistence related things.
"""

import asyncio
from concurrent.futures import ThreadPoolExecutor
from collections import defaultdict
import importlib

from sqlitedict import SqliteDict

from pypeman.conf import settings

SENTINEL = object()

_backend = None


async def get_backend(loop):
    """
    Return the configured backend instance.

    :param loop: Asyncio loop to use. Passed backend instance.
    """
    global _backend
    if not _backend:
        # Load backend on first use
        if not settings.PERSISTENCE_BACKEND:
            raise Exception("Persistence backend not configured.")

        module, _, class_ = settings.PERSISTENCE_BACKEND.rpartition('.')
        loaded_module = importlib.import_module(module)
        _backend = getattr(loaded_module, class_)(loop=loop, **settings.PERSISTENCE_CONFIG)

        await _backend.start()

    return _backend


class MemoryBackend():
    """
    Memory persistence backend.
    Only for testing purpose as this is not really persistent.
    """
    def __init__(self, loop=None):
        self.loop = loop
        self._data = defaultdict(dict)

    async def start(self):
        """
        Do nothing for this backend.
        """
        pass

    async def store(self, namespace, key, value):
        """ Store the value in a dict in memory

        :param namespace: Namespace for dict.
        :param key: Access key.
        :param value: Value to save.
        """
        self._data[namespace][key] = value

    async def get(self, namespace, key, default=SENTINEL):
        """
        Get the value from memory.

        :param namespace: Namespace for dict.
        :param key: Key to get.
        :param default: Default value if key missing.
        """
        if default is not SENTINEL:
            return self._data[namespace].get(key, default)
        else:
            return self._data[namespace][key]


class SqliteBackend():
    """
    Sqlite persistence backend. Store data in an sqlite database with
    ACID garanties. Internally use a thread pool to execute database access.

    :param path: Path of sqlite file.
    :param thread_pool: If you want a specific thread_pool you can give one here.
    :param loop: Loop used for the executor.
    """
    def __init__(self, path, thread_pool=None, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self.executor = thread_pool or ThreadPoolExecutor(max_workers=1)
        self.path = path

    async def start(self):
        """
        Do nothing for this backend.
        """
        pass

    def _sync_store(self, namespace, key, value):
        with SqliteDict(self.path, tablename=namespace) as pdict:
            pdict[key] = value
            pdict.commit()

    def _sync_get(self, namespace, key, default):
        with SqliteDict(self.path, tablename=namespace) as pdict:
            if default is not SENTINEL:
                return pdict.get(key, default)
            else:
                return pdict[key]

    async def store(self, namespace, key, value):
        """ Store the value in a dict saved in sqlite db.

        :param namespace: Namespace table name.
        :param key: Access key.
        :param value: Value to save.
        """
        await self.loop.run_in_executor(self.executor, self._sync_store, namespace, key, value)

    async def get(self, namespace, key, default=SENTINEL):
        """
        Get the value from sqlite db.

        :param namespace: Namespace table name.
        :param key: Key to get.
        :param default: Default value if key missing.
        """
        return await self.loop.run_in_executor(self.executor, self._sync_get, namespace, key, default)



import datetime
import dateutil.parser
import logging
import os
import re

from itertools import islice
from collections import OrderedDict

from pypeman.message import Message

from pypeman.errors import PypemanConfigError

logger = logging.getLogger("pypeman.store")

DATE_FORMAT = '%Y%m%d_%H%M'


class MessageStoreFactory():
    """ Message store factory class can generate Message store instance for specific store_id. """

    def get_store(self, store_id):
        """
        :param store_id: identifier of corresponding message store.
        :return: A MessageStore corresponding to correct store_id.
        """


class MessageStore():
    """ A MessageStore keep an history of processed messages. Mainly used in channels. """

    async def start(self):
        """
        Called at startup to initialize store.
        """

    async def store(self, msg):
        """
        Store a message in the store.

        :param msg: The message to store.
        :return: Id for this specific message.
        """

    async def change_message_state(self, id, new_state):
        """
        Change the `id` message state.

        :param id: Message specific store id.
        :param new_state: Target state.
        """

    async def get(self, id):
        """
        Return one message corresponding to given `id` with his status.

        :param id: Message id. Message store dependant.
        :return: A dict `{'id':<message_id>, 'state': <message_state>, 'message': <message_object>}`.
        """

    async def get_preview_str(self, id):
        """
        Return the first 1000 chars of message content corresponding to `id`

        :param id: Message id. Message store dependant.
        :return: A dict `{'id':<message_id>, 'message_content': str}`.
        """

    async def get_msg_content(self, id):
        """
        Return the content of the message corresponding to given `id`

        :param id: Message id. Message store dependant.
        :return: A dict `{'id':<message_id>, 'message_content': Message}`.
        """

    async def is_regex_in_msg(self, id, rtext):
        """
        Return True if the str(msg) contains the regex rtext

        :param id: Message id. Message store dependant.
        :param rtext: string of regular expression to search in msg
        :return: True if it matches False otherwise
        """

    async def is_txt_in_msg(self, id, text):
        """
        Return True if the str(msg) contains param text

        :param id: Message id. Message store dependant.
        :param text: String. The text to search in msg
        :return: True if it text is found, False otherwise
        """

    async def search(self, start=0, count=10, order_by='timestamp', start_dt=None, end_dt=None,
                     text=None, rtext=None):
        """
        Return a list of message with store specific `id` and processed status.

        :param start: First element.
        :param count: Count of elements since first element.
        :param order_by: Message order. Allowed values : ['timestamp', 'status'].
        :param start_dt: (optional) Isoformat start date(time) to filter with
        :param end_dt: (optional) Isoformat end date(time) to filter with
        :param text: (optional) String to search in message content
        :param rtext: (optional) String regex to search in message content
        :return: A list of dict `{'id':<message_id>, 'state': <message_state>, 'message': <message_object>}`.
        """

    async def total(self):
        """
        :return: total count of messages
        """


class NullMessageStoreFactory(MessageStoreFactory):
    """ Return an NullMessageStore that do nothing at all. """
    def get_store(self, store_id):
        return NullMessageStore()


class NullMessageStore(MessageStore):
    """ For testing purpose """

    async def store(self, msg):
        return None

    async def get(self, id):
        return None

    async def get_preview_str(self, id):
        return None

    async def get_msg_content(self, id):
        return None

    async def search(self, **kwargs):
        return None

    async def total(self):
        return 0


class FakeMessageStoreFactory(MessageStoreFactory):
    """ Return an Fake message store """
    def get_store(self, store_id):
        return FakeMessageStore()


class FakeMessageStore(MessageStore):
    """ For testing purpose """

    async def store(self, msg):
        logger.debug("Should store message %s", msg)
        return 'fake_id'

    async def get(self, id):
        return {'id': id, 'state': 'processed', 'message': None}

    async def get_preview_str(self, id):
        return {"id": id, "message_content": "content"}

    async def get_msg_content(self, id):
        return {"id": id, "message_content": "content"}

    async def is_regex_in_msg(self, id, rtext):
        return True

    async def is_txt_in_msg(self, id, text):
        return True

    async def search(self, **kwargs):
        return []

    async def total(self):
        return 0


class MemoryMessageStoreFactory(MessageStoreFactory):
    """ Return a Memory message store. All message are lost at pypeman stop. """
    def __init__(self):
        self.base_dict = {}

    def get_store(self, store_id):
        return MemoryMessageStore(self.base_dict, store_id)


class MemoryMessageStore(MessageStore):
    """ Store messages in memory """

    def __init__(self, base_dict, store_id):
        super().__init__()
        self.messages = base_dict.setdefault(store_id, OrderedDict())

    async def store(self, msg):
        msg_id = msg.uuid
        self.messages[msg_id] = {
            'id': msg_id, 'state': Message.PENDING,
            'timestamp': msg.timestamp, 'message': msg.to_dict()}
        return msg_id

    async def change_message_state(self, id, new_state):
        self.messages[id]['state'] = new_state

    async def get(self, id):
        resp = dict(self.messages[id])
        resp['message'] = Message.from_dict(resp['message'])
        return resp

    async def get_preview_str(self, id):
        msg = await self.get_msg_content(id)
        try:
            msg.payload = str(msg.payload)[:1000]
        except Exception:
            msg.payload = repr(msg.payload)[:1000]
        return msg

    async def get_msg_content(self, id):
        msg = await self.get(id)
        msg_content = msg["message"]
        return msg_content

    async def is_regex_in_msg(self, id, rtext):
        msg = await self.get_msg_content(id)
        try:
            msg.payload = str(msg.payload)
        except Exception:
            msg.payload = repr(msg.payload)
        regex = re.compile(rtext)
        return True if regex.match(msg.payload) else False

    async def is_txt_in_msg(self, id, text):
        msg = await self.get_msg_content(id)
        try:
            msg.payload = str(msg.payload)
        except Exception:
            msg.payload = repr(msg.payload)
        return text in msg.payload

    async def search(self, start=0, count=10, order_by='timestamp', start_dt=None, end_dt=None,
                     text=None, rtext=None):

        if order_by.startswith('-'):
            reverse = True
            sort_key = order_by[1:]
        else:
            reverse = False
            sort_key = order_by

        if start_dt:
            start_dt = dateutil.parser.isoparse(start_dt)
        if end_dt:
            end_dt = dateutil.parser.isoparse(end_dt)

        result = []
        values = (
            val for val in self.messages.values()
            if (not start_dt or val["timestamp"] >= start_dt)
            and (not end_dt or val["timestamp"] <= end_dt)
        )
        if text:
            found_values = []
            for val in values:
                if await self.is_txt_in_msg(val["id"], text):
                    found_values.append(val)
            values = found_values

        if rtext:
            found_values = []
            for val in values:
                if await self.is_regex_in_msg(val["id"], rtext):
                    found_values.append(val)
            values = found_values
        for value in islice(sorted(values, key=lambda x: x[sort_key], reverse=reverse),
                            start, start + count):
            resp = dict(value)
            resp['message'] = Message.from_dict(resp['message'])
            result.append(resp)
        return result

    async def total(self):
        return len(self.messages)


class FileMessageStoreFactory(MessageStoreFactory):
    """
    Generate a FileMessageStore message store instance.
    Store a file in `<base_path>/<store_id>/<month>/<day>/` hierachy.
    """

    # TODO add an option to reguraly archive old file or delete them
    def __init__(self, path):
        super().__init__()
        if path is None:
            raise PypemanConfigError('file message store requires a path')
        self.base_path = path

    def get_store(self, store_id):
        return FileMessageStore(self.base_path, store_id)


class FileMessageStore(MessageStore):
    """ Store a file in `<base_path>/<store_id>/<month>/<day>/` hierachy."""
    # TODO file access should be done in another thread. Waiting for file backend.

    def __init__(self, path, store_id):
        super().__init__()

        self.base_path = os.path.join(path, store_id)

        # Match msg file name
        self.msg_re = re.compile(
            r'^(?P<msg_date>[0-9]{8})_(?P<msg_time>[0-9]{2}[0-9]{2})_(?P<msg_uid>[0-9abcdef]*)$')

        try:
            # Try to make dirs if necessary
            os.makedirs(os.path.join(self.base_path))
        except FileExistsError:
            pass

        self._total = 0

    async def start(self):
        self._total = await self.count_msgs()

    async def store(self, msg):
        """ Store a file in `<base_path>/<store_id>/<month>/<day>/` hierachy."""
        # TODO implement a safer store to avoid broken messages

        # The filename is the file id
        filename = "{}_{}".format(msg.timestamp.strftime(DATE_FORMAT), msg.uuid)
        dirs = os.path.join(str(msg.timestamp.year),
                            "%02d" % msg.timestamp.month,
                            "%02d" % msg.timestamp.day)

        try:
            # Try to make dirs if necessary
            os.makedirs(os.path.join(self.base_path, dirs))
        except FileExistsError:
            pass

        file_path = os.path.join(dirs, filename)

        # Write message to file
        with open(os.path.join(self.base_path, file_path), "w") as f:
            f.write(msg.to_json())

        await self.change_message_state(file_path, Message.PENDING)

        self._total += 1

        return file_path

    async def change_message_state(self, id, new_state):
        with open(os.path.join(self.base_path, id + '.meta'), "w") as f:
            f.write(new_state)

    async def get_message_state(self, id):
        with open(os.path.join(self.base_path, id + '.meta'), "r") as f:
            state = f.read()
            return state

    async def get(self, id):
        if not os.path.exists(os.path.join(self.base_path, id)):
            raise IndexError

        with open(os.path.join(self.base_path, id), "rb") as f:
            msg = Message.from_json(f.read().decode('utf-8'))
            return {'id': id, 'state': await self.get_message_state(id), 'message': msg}

    async def get_preview_str(self, id):
        msg = await self.get_msg_content(id)
        try:
            msg.payload = str(msg.payload)[:1000]
        except Exception:
            msg.payload = repr(msg.payload)[:1000]
        return msg

    async def get_msg_content(self, id):
        msg = await self.get(id)
        msg_content = msg["message"]
        return msg_content

    async def sorted_list_directories(self, path, reverse=True):
        """
        :param path: Base path
        :param reverse: reverse order
        :return: List of directories in specified path ordered
        """
        return sorted([d for d in os.listdir(path) if os.path.isdir(os.path.join(path, d))], reverse=reverse)

    async def count_msgs(self):
        """
        Count message by listing all directories. To be used at startup.
        """
        count = 0
        for year in await self.sorted_list_directories(os.path.join(self.base_path)):
            for month in await self.sorted_list_directories(os.path.join(self.base_path, year)):
                for day in await self.sorted_list_directories(os.path.join(self.base_path, year, month)):
                    for msg_name in sorted(os.listdir(os.path.join(self.base_path, year, month, day))):
                        found = self.msg_re.match(msg_name)
                        if found:
                            count += 1
        return count

    async def is_regex_in_msg(self, id, rtext):
        msg = await self.get_msg_content(id)
        try:
            msg.payload = str(msg.payload)
        except Exception:
            msg.payload = repr(msg.payload)
        regex = re.compile(rtext)
        return True if regex.match(msg.payload) else False

    async def is_txt_in_msg(self, id, text):
        msg = await self.get_msg_content(id)
        try:
            msg.payload = str(msg.payload)
        except Exception:
            msg.payload = repr(msg.payload)
        return text in msg.payload

    async def search(self, start=0, count=10, order_by='timestamp', start_dt=None, end_dt=None,
                     text=None, rtext=None):
        # TODO better performance for slicing by counting file in dirs ?
        if order_by.startswith('-'):
            reverse = True
            # sort_key = order_by[1:]
        else:
            reverse = False
            # sort_key = order_by

        if start_dt:
            start_dt = dateutil.parser.isoparse(start_dt)
        if end_dt:
            end_dt = dateutil.parser.isoparse(end_dt)

        # TODO handle sort_key
        result = []
        end = start + count
        position = 0
        for year in await self.sorted_list_directories(
                os.path.join(self.base_path), reverse=reverse):
            for month in await self.sorted_list_directories(
                    os.path.join(self.base_path, year), reverse=reverse):
                for day in await self.sorted_list_directories(
                        os.path.join(self.base_path, year, month), reverse=reverse):
                    # Pre filter with date to avoid listing unuseful dirs
                    msg_date = datetime.date(year=int(year), month=int(month), day=int(day))
                    if start_dt:
                        if msg_date < start_dt.date():
                            continue
                    if end_dt:
                        if msg_date > end_dt.date():
                            continue
                    for msg_name in sorted(
                            os.listdir(os.path.join(
                                self.base_path, year, month, day)),
                            reverse=reverse):
                        found = self.msg_re.match(msg_name)
                        if found:
                            msg_str_time = found.groupdict()["msg_time"]
                            hour = int(msg_str_time[:2])
                            minute = int(msg_str_time[2:4])
                            msg_time = datetime.time(hour=hour, minute=minute, second=0)
                            msg_dt = datetime.datetime.combine(msg_date, msg_time)
                            mid = os.path.join(year, month, day, msg_name)
                            if start_dt:
                                if msg_dt < start_dt:
                                    continue
                            if end_dt:
                                if msg_dt > end_dt:
                                    continue
                            if text:
                                if not await self.is_txt_in_msg(mid, text):
                                    continue
                            if rtext:
                                if not await self.is_regex_in_msg(mid, rtext):
                                    continue
                            if start <= position < end:
                                result.append(await self.get(mid))
                            position += 1
        return result

    async def total(self):
        return self._total
